#  Copyright 2021 Technische Universiteit Delft
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import numpy as np
from random import uniform

# Create VTT midpoints
def vtt_midpoints(vtt_grid):
    
    # Calculate midpoints
    vtt_grid_midpoints = ((np.append(vtt_grid, 0) + np.append(0,vtt_grid)))/2
    
    # Replace last value of the VTT midpoints by the last value of the VTT array
    vtt_grid_midpoints[-1] = vtt_grid[-1]

    return vtt_grid_midpoints

# Create predicted VTT for each respondent, based in the choice probability and a VTT grid.
def predicted_vtt(ecdf,grid,NP):
    pp = ((np.append(ecdf, 0)-np.append(0,ecdf)))
    count_data = []
    for n in range(1,len(grid)):
        dat = [uniform(grid[n-1], grid[n]) for p in range(0, np.round((pp[n]*NP)).astype(int))]
        count_data = np.append(count_data,dat)
    return count_data

"""BFGSMin functions"""
# BFGSMin functions
# Written by Jose Ignacio Hernandez

# Import packages
import numpy as np

# BFGS Minimizer function
def _bfgsmin(f,x0,maxiter=1000,tol=np.sqrt(np.finfo(float).eps),verbose=False,difftype='central',diffeps=np.sqrt(np.finfo(float).eps),steptol=1e-30,args=()):


    # Ignore warnings
    import warnings
    warnings.filterwarnings("ignore")

    # Initialize parameters
    x = np.array(x0)                            # Initial value of x
    f_val = f(x,*args)                          # Initial value for objective function
    g0 = _numgr(f,x,difftype,diffeps,*args)      # Initial value of gradient
    H0 = np.eye(x.shape[0])                     # Initial value of approximate Hessian is a positive definite matrix (e.g. identity matrix)
    g_diff = np.inf                             # Initial difference in gradient norm is set to infinity
    c1 = 1e-4                                   # Internal scalar for the Armijo-Goldstein condition (for step size computation)
    lambd = 1                                   # Initial value of the step size
    convergence = -1                            # Set convergence flag to -1. If zero, the algorithm converged

    # Print inital value of the objective function
    if verbose:
        print('Initial F-value: ' + str(round(f_val,2)))

    # Start algorithm
    for iter in range(maxiter):
        
        # If gradient norm is less than tolerance value, convergence is achieved and the loop is broken
        if g_diff < tol:
            convergence = 0

            if verbose:
                print('\nLocal minimum found. G-norm below tolerance')

            break

        # Set step size = 1 on each new iteration
        lambd = 1

        # Construct direction vector and relative gradient

        # d = np.linalg.solve(-H0,np.eye(len(x))) @ g0
        d = np.linalg.inv(-H0) @ g0
        m = d.T @ g0
        # Select step size that satisfies the Armijo-Goldstein condition
        while True:
            
            # If lambd decreases less than tol, then stop and return convergence = 3
            if lambd < steptol:
                if verbose:
                    print('\nLocal minimum possible. Step size tolerance limit reached.')
                convergence = 5
                break

            # Evaluate objective function using current step size
            x1 = x + lambd*d
            f1 = f(x1,*args)

            # Construct test for A-G condition
            ftest = f_val + c1*lambd*m

            # If A-G condition is satisfied (i.e current step size generates a decrease of the objective function), continue with BFGS algorithm
            if (np.isnan(f1) == False) and (f1 <= ftest) and (f1 > 0):
                break

            # ...else, decrease step size
            else:
                lambd = lambd/2

        # If step size calculation loop was broken, then break the algorithm and return convergence code + partial results
        if convergence == 5:
            break

        # BFGS ALGORITHM: construct the improvement and gradient improvement
        g1 = _numgr(f,x1,difftype,diffeps,*args)
        s0 = (lambd*d)[:,np.newaxis]
        y0 = (g1 - g0)[:,np.newaxis]

        # Update Hessian using BFGS formula
        H0 = H0 + (y0 @ y0.T) / (y0.T @ s0) - (((H0 @ s0) @ s0.T) @ H0)/((s0.T @ H0) @ s0)
        
        # Store new gradient and compute the new value of objective function
        x = x1.copy()
        g0 = g1.copy()
        f_val = f1.copy()

        # Compute infinite norm of gradient vector 
        g_diff = np.abs(m)
        # Print output
        if verbose:
            print('Iter No. ' + str(int(iter+1)) + ': F-value: ' + str(round(f_val,2)) + ' / Step size: ' + str(round(lambd,6)) + ' / G-norm: ' + str(round(g_diff,6)))

    # If the algorithm reaches the maximum iterations, it prints convergence message = 2
    if iter == maxiter:
        convergence = 2

    # Return the Hessian approximation
    H = H0

    # Return convergence flag, iterations, final f value, final x value, and final approx. hessian
    return({'convergence': convergence, 'iterations': iter+1, 'fun': f_val, 'x': x, 'hessian': H})

# Numeric gradient fuction
def _numgr(f,param,difftype='forward',eps=np.sqrt(np.finfo(float).eps),*args):    
    
    # Define scalars and initialize vectors
    K = len(param)                          # No. of parameters
    gr = np.full(K,np.nan)                  # Initialize gradient vector
    ej = np.eye(K)*eps                      # Vector of eps
    
    # If difftype == 'central', then:
    if difftype == 'central':
        for k in range(K):
            gr[k] = (f(param + ej[:,k],*args) - f(param - ej[:,k],*args))*0.5/eps

    # ...else, if difftype == 'forward' (default):
    elif difftype == 'forward':
        f0 = f(param,*args)
        for k in range(K):
            gr[k] = (f(param + ej[:,k],*args) - f0)/eps

    # ...else, return error
    else:
        raise ValueError('''difftype must be either 'forward' or 'central' ''')

    # Return gradient vector
    return(gr)

# Numeric Hessian function
class numhess:
    
    def __init__(self,f,eps=np.sqrt(np.finfo(float).eps)):
        self.f = f
        self.eps = eps
    
    def __call__(self,param,*args):
        eps = self.eps
        f = self.f

        # Define scalars and initialize vectors
        K = len(param)                          # No. of parameters
        hs = np.full((K,K),np.nan)              # Initialize Hessian vector
        ej = np.eye(K)*eps                      # Vector of eps

        # f0 = f(param,*args)
        for i in range(K):
            for j in range(K):
                # f1 = f(param + ej[:,i] + ej[:,j],*args)
                # f2 = f(param + ej[:,i],*args)
                # f3 = f(param + ej[:,j],*args)

                # hs[i,j] = (f1-f2-f3+f0)/(eps**2)
                # if i != j:
                #     hs[i,j] = (f1-f2-f3+f0)/(eps**2)
                
                f1 = f(param + ej[:,i] + ej[:,j],*args)
                f2 = f(param + ej[:,i] - ej[:,j],*args)
                f3 = f(param - ej[:,i] + ej[:,j],*args)
                f4 = f(param - ej[:,i] - ej[:,j],*args)

                hs[i,j] = (f1-f2-f3+f4)/(4*eps*eps)
                if i != j:
                    hs[j,i] = (f1-f2-f3+f4)/(4*eps*eps)
    
        return hs



#  Copyright 2021 Technische Universiteit Delft
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""Modules to configure and estimate a Rouwendal model."""
from dataclasses import dataclass
from typing import Tuple
from numdifftools import Hessian
import numpy as np
import warnings
from py_np4vtt.data_format import ModelArrays
from py_np4vtt.utils import vtt_midpoints, predicted_vtt,_bfgsmin
import time

warnings.filterwarnings('ignore')

@dataclass
class ConfigRouwendal:
    """Configuration class of the Rouwendal model.
    
    This class stores the configuration parameters of a Rouwendal model 
    and performs integrity checks before being passed to the model object.
    
    Parameters
    ----------
    
    minimum : float
        Minimum value of the VTT grid
    maximum : float
        Maximum value of the VTT grid
    supportPoints : int
        Number of support points of the VTT grid. The VTT grid will contain
        `(supportPoints-1)` intervals. Must be greater than zero
    startQ : float
        Starting value of the probability of consistent choice. Must be
        between zero and one.
    """
    minimum: float
    maximum: float
    supportPoints: int
    startQ: float

    def validate(self):
        # Create errormessage list
        errorList = []

        if not self.maximum > self.minimum:
            errorList.append('Max must be greater than minimum.')

        if not self.supportPoints > 0:
            errorList.append('No. of support points must be greater than zero.')

        if not (0 < self.startQ < 1):
            errorList.append('Probability of consistent choice must be in the interval (0,1).')

        # Whoever calls this validator knows that empty errorList means validator success
        return errorList

class ModelRouwendal:
    """Rouwendal model.
    
    This is the model class that prepares the data and estimates 
    the Rouwendal model [1].
    
    Parameters
    -----------
    params : ConfigRouwendal
        A configuration class of a Rouwendal model.
    arrays : ModelArrays
        Model arrays created with `make_modelarrays`
        
    Attributes
    ----------
    vtt_grid : numpy.ndarray
        The VTT grid created with the specifications of `ConfigRouwendal`.
    vtt_mid : numpy.ndarray
        The mid points of the VTT grid.

    Methods
    -------
    run():
        Estimates the Rouwendal model.
    
    References
    ----------
    [1] Rouwendal, Jan, et al. "The information content of a stated choice 
    experiment: A new method and its application to the value of a 
    statistical life." Transportation research part B: methodological 44.1 
    (2010): 136-151.
    """
    def __init__(self, cfg: ConfigRouwendal, arrays: ModelArrays):
        self.cfg = cfg
        self.arrays = arrays

        # Check if data is a balanced panel. Otherwise raise an error
        assert arrays.is_balanced_panel, "Data is not a balanced panel. ModelRouwendal is only compatible with balanced panel data"

        # Create grid of support points
        self.vtt_grid = np.linspace(self.cfg.minimum, self.cfg.maximum, self.cfg.supportPoints)

        # Compute the midpoints of the VTT grid
        self.vtt_mid = vtt_midpoints(self.vtt_grid)

        # Compute distance between points at each support point
        dist = self.vtt_grid[1] - self.vtt_grid[0]

        # Print message of the support points
        print("Created a VTT grid of " + str(self.cfg.supportPoints) + \
            " points between " + str(self.cfg.minimum) + " and " + str(self.cfg.maximum) + ".")

        print("Distance between points of the VTT grid is " + str(dist))
        
    def run(self):
        """Estimates the Rouwendal model.
        
        Parameters
        ----------
        None.

        Returns
        -------
        q_est : float
            The estimated parameter that generates the probability of 
            consistent choice.
        q_se : float
            The standard error of the parameter of consistent choice.
        q_prob : float
            The estimated probability of consistent choice.
        x : numpy.ndarray
            The estimated density parameters at each point of the VTT grid.
        se : numpy.ndarray
            The standard errors of `x`.
        p : numpy.ndarray
            The estimates of the cumulative probability function (CDF) of 
            the VTT at each support point. The first point is always zero 
            for compatibility with plots.
        vtt : numpy.ndarray
            The estimated VTT for each respondent, based in the estimated 
            CDF poonts (p) and the VTT grid.
        init_ll : float
            Value of log-likelihood function in the initial value of startQ. 
            Starting values of support point parameters are equal to zero.
        ll : float
            Value of the likelihood function in the optimum.
        exitflag : int
            Exit flag of the optimisation routine. If `exitflag=0`, the 
            optimisation succeeded. Otherwise, check the configuration parameters.
        est_time : float
            The estimation time in seconds.
        """
        # Set vector of starting values of xameters to estimate
        q0 = np.log(self.cfg.startQ/(1-self.cfg.startQ))
        x0 = np.hstack([q0, np.zeros(len(self.vtt_grid))])

        # Get VTT grid shape
        grid_shape = len(self.vtt_grid)

        # Prepare arrays
        BVTT_array = np.tile(self.arrays.BVTT,(grid_shape,1,1)).T
        Choice_array = np.tile(self.arrays.Choice,(grid_shape,1,1)).T

        # Initial value of the log-likelihood function
        init_ll = -ModelRouwendal.objectiveFunction(x0, self.arrays.NP, self.arrays.T, BVTT_array,
                                                      Choice_array, self.vtt_grid)

        # TODO: add an integrity check: initialVal should be finite. Otherwise, rise an error.

        # Starting values
        argTuple = (self.arrays.NP, self.arrays.T, BVTT_array, Choice_array, self.vtt_grid)

        # Start optimization
        t0 = time.time()
        results = _bfgsmin(ModelRouwendal.objectiveFunction, x0, args=argTuple,tol=1e-6,verbose=True)

        # Collect results
        x = results['x']
        hess = Hessian(ModelRouwendal.objectiveFunction,method='forward')(x,self.arrays.NP, self.arrays.T, BVTT_array, Choice_array, self.vtt_grid)
        se = np.sqrt(np.diag(np.linalg.inv(hess)))
        ll = -results['fun']
        exitflag = results['convergence']

        # Compute elapsed time
        t1 = time.time()
        est_time = t1 - t0
        
        # Get estimated probability of consistent choice
        q_prob = np.exp(x[0])/(1+np.exp(x[0]))
        q_est = x[0]
        q_se = np.sqrt((np.exp(q_est)/(1+np.exp(q_est)))**2)**2 * se[0]**2

        # Get estimated FVTT and xameters
        x = x[1:]
        se = se[1:]
        fvtt = np.exp(x)/np.sum(np.exp(x))
        p = np.cumsum(fvtt)

        # Compute the predicted VTT at the midpoints
        vtt = predicted_vtt(p,self.vtt_mid,self.arrays.NP)

        # Add point 0 in the estimated CDF to make coincide with point zero in the VTT mid point
        p = np.concatenate((0,p),axis=None)

        # Return output
        return q_est, q_se, q_prob, x, se, p, vtt, init_ll, ll, exitflag, est_time

    @staticmethod
    def objectiveFunction(x, NP, T, BVTT, Choice, vtt_grid):
        
        # Re-scale Q and FVTT to fit between zero and one
        q = np.exp(x[0]) / (1 + np.exp(x[0]))
        fvtt = np.exp(x[1:]) / np.sum(np.exp(x[1:]))

        # Compute conditional probabilities
        tau = ((vtt_grid > BVTT) == Choice).astype(int).sum(axis=0)
        P = ((q**tau) * (1-q)**(T-tau))

        # Conditional probability
        # P = np.zeros((NP, len(vtt_grid)))

        # for n in range(len(vtt_grid)):
        #     tau = np.sum(((vtt_grid[n] > BVTT) == Choice).astype(int), axis=1)
        #     P[:, n] = (q**tau) * ((1-q) ** (T-tau))
            
        # Maximise log-likelihood. L is computed by multiplying conditional P
        # with density fvtt, average and sum across all obs
        L = -np.sum(np.log(np.sum(fvtt*P, axis=1)))

        return L



#  Copyright 2021 Technische Universiteit Delft
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from typing import Optional
from enum import Enum, unique, auto
from math import ceil, floor, sqrt

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator


# Histogram can only be produced for Logit and ANN
def run_plot_hist(vtts: np.ndarray,
        min_bvtt_arg: Optional[float] = None, max_bvtt_arg: Optional[float] = None,
        bin_width_arg: Optional[float] = None):

    # Set min, max and bin width for automatic mode
    auto_min_bvtt = floor(np.min(vtts));
    auto_max_bvtt = ceil(np.max(vtts));
    min_bvtt = auto_min_bvtt if min_bvtt_arg is None else min_bvtt_arg
    max_bvtt = auto_max_bvtt if max_bvtt_arg is None else max_bvtt_arg

    spec_tolerance = max_bvtt - min_bvtt;

    N = vtts.size()
    n_bins = ceil(sqrt(N));

    auto_bin_width = round(floor(spec_tolerance) / n_bins, 2);
    bin_width = auto_bin_width if bin_width_arg is None else bin_width_arg

    pass


@unique
class PlotMethodLogitOrANN(Enum):
    LOGIT = auto()
    ANN = auto()

    def __str__(self):
        if self.value == PlotMethodLogitOrANN.LOGIT: return 'Logit'
        else: return 'ANN'


def run_plot_ecdf_logit_ann(
        plot_method: PlotMethodLogitOrANN,
        vtts: np.ndarray, probs: np.ndarray,
        min_bvtt: Optional[float] = None, max_bvtt: Optional[float] = None,
        ):
    # Grid of subfigures
    fig = plt.figure()
    axs = fig.add_subplot(1, 1, 1)  # Single subfigure

    # Decorations
    axs.set_title(f"{plot_method} ECDF")
    axs.set_xlabel('VTT')
    axs.set_ylabel('Cumulative density')

    # This is because the cumulative prob. when x=0 should be 0.0
    data_x = np.append([0.0], vtts[1:])
    data_y = np.append([0.0], probs)

    # The point (0,0) should be exactly on the bottom left, no margins
    axs.margins(0.0)

    # formatting just a parameter or...?
    format_points_ecdf = 'bo-'
    axs.plot(data_x, data_y, format_points_ecdf)

    # Axis limits and ticks
    ylim_min = 0.0   # Must be like this because it's a probabilty
    ylim_max = 1.0
    yticks_num = 11
    axs.set_xlim(left=min_bvtt, right=max_bvtt)
    axs.set_ylim(bottom=ylim_min, top=ylim_max)
    axs.set_yticks(np.linspace(ylim_min, ylim_max, yticks_num))

    # TODO: check that matplotlib is smart to not show ALL THE POINTS
    axs.xaxis.set_major_locator(MaxNLocator(integer=True))

    plt.show()


@unique
class PlotMethodRWOrLL(Enum):
    ROUWENDAL = auto()
    LOC_LOGIT = auto()

    def __str__(self):
        if self.value == PlotMethodRWOrLL.ROUWENDAL: return 'Rouwendal'
        else: return 'Local logit'


def run_plot_ecdf_rouwendal_loclogit(
        plot_method: PlotMethodRWOrLL,
        vtts: np.ndarray, probs: np.ndarray,
        min_bvtt: Optional[float] = None, max_bvtt: Optional[float] = None,
        ):
    # Grid of subfigures
    fig = plt.figure()
    axs = fig.add_subplot(1, 1, 1)  # Single subfigure

    # Decorations
    axs.set_title(f"{plot_method} ECDF")
    axs.set_xlabel('VTT')
    axs.set_ylabel('Cumulative density')

    # This is because the cumulative prob. when x=0 should be 0.0
    data_x = np.append([0.0], vtts[1:])
    data_y = np.append([0.0], probs)

    # The point (0,0) should be exactly on the bottom left, no margins
    axs.margins(0.0)

    # formatting just a parameter or...?
    format_points_ecdf = 'bo-'
    axs.plot(data_x, data_y, format_points_ecdf)

    # Axis limits and ticks
    ylim_min = 0.0   # Must be like this because it's a probabilty
    ylim_max = 1.0
    yticks_num = 11
    axs.set_xlim(left=min_bvtt, right=max_bvtt)
    axs.set_ylim(bottom=ylim_min, top=ylim_max)
    axs.set_yticks(np.linspace(ylim_min, ylim_max, yticks_num))

    # TODO: check that matplotlib is smart to not show ALL THE POINTS
    axs.xaxis.set_major_locator(MaxNLocator(integer=True))

    plt.show()




#  Copyright 2021 Technische Universiteit Delft
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""Modules to configure and estimate a Local constant model."""

from dataclasses import dataclass
import numpy as np
from scipy.stats import norm
from py_np4vtt.data_format import ModelArrays
from py_np4vtt.utils import predicted_vtt, vtt_midpoints
import time

@dataclass
class ConfigLConstant:
    """Configuration class of the local constant model.
    
    This class stores the configuration parameters of a local constant model 
    and performs integrity checks before being passed to the model object.
    
    Parameters
    ----------
    
    minimum : float
        Minimum value of the VTT grid
    maximum : float
        Maximum value of the VTT grid
    supportPoints : int
        Number of support points of the VTT grid. The VTT grid will contain 
        `(supportPoints-1)` intervals. Must be greater than zero
    kernelWidth : float
        Kernel width for the Nadaraya-Watson estimator. Must be greater than 
        zero.
    """
    minimum: float
    maximum: float
    supportPoints: int
    kernelWidth: float

    def validate(self):
        # Create errormessage list
        errorList = []

        if not self.maximum > self.minimum:
            errorList.append('Max must be greater than minimum.')

        if not self.supportPoints > 0:
            errorList.append('No. of support points must be greater than zero.')

        if not self.kernelWidth > 0:
            errorList.append('Kernel width must be greater than zero.')

        # Whoever calls this validator knows that empty errorList means validator success
        return errorList

class ModelLConstant:
    """Local constant model.
    
    This is the model class that prepares the data and estimates a local
    constant model [1,2].
    
    Parameters
    -----------
    params : ConfigLConstant
        A configuration class of a local constant model.
    arrays : ModelArrays
        Model arrays created with `make_modelarrays`
        
    Attributes
    ----------
    vtt_grid : numpy.ndarray
        The VTT grid created with the specifications of `ConfigLConstant`
    vtt_mid : numpy.ndarray
        The mid points of the VTT grid.

    Methods
    -------
    run():
        Estimates the local constant model.
    
    References
    ----------
    [1] Fosgerau, Mogens. "Investigating the distribution of the value of 
    travel time savings." Transportation Research Part B: Methodological 
    40.8 (2006): 688-707.
    [2] Fosgerau, Mogens. "Using nonparametrics to specify a model to measure 
    the value of travel time." Transportation Research Part A: Policy and 
    Practice 41.9 (2007): 842-856.
    """
    def __init__(self, params: ConfigLConstant, arrays: ModelArrays):
        self.params = params
        self.arrays = arrays

        # Create grid of support points
        self.vtt_grid = np.linspace(self.params.minimum, self.params.maximum, self.params.supportPoints)

        # Compute the midpoints of the VTT grid
        self.vtt_mid = vtt_midpoints(self.vtt_grid)

        # Compute distance between points at each support point
        dist = self.vtt_grid[1] - self.vtt_grid[0]

        # Print message of the support points
        print("Created a VTT grid of " + str(self.params.supportPoints) + \
            " points between " + str(self.params.minimum) + " and " + str(self.params.maximum) + ".")

        print("Distance between points of the VTT grid is " + str(dist))

    def run(self):
        """Estimates the local constant model.
        
        Parameters
        ----------
        None.

        Returns
        -------
        p : numpy.ndarray
            The estimates of the cumulative choice probability (CDF), evaluated 
            at each mid point of the VTT grid. The first point is always zero 
            while the last point is always equal to the second last for 
            compatibility with plots.
        vtt : numpy.ndarray
            The estimated VTT per respondent, based in the estimated 
            CDF points (`p`) and the sample.
        est_time : float
            The estimation time in seconds.
        """

        # Start estimation
        t0 = time.time()
        p = ModelLConstant.nadaraya_watson(self.vtt_mid[1:-1],~self.arrays.Choice.flatten(),self.arrays.BVTT.flatten(),self.params.kernelWidth)

        # Create counts per point of the VTT mid points
        vtt = predicted_vtt(p,self.vtt_grid,self.arrays.NP)

        # Compute elapsed time
        t1 = time.time()
        est_time = t1 - t0

        # Add point 0 in the estimated CDF and repeat last point to make coincide with point zero and last point in the VTT mid point
        p = np.concatenate((0,p,p[-1]),axis=None)

        # Return list of outcomes
        return p, vtt, est_time

    # Nadaraya-Watson estimator with gaussian kernel
    @staticmethod
    def nadaraya_watson(x,Y,X,h):
        
        g = np.empty(shape=x.shape)

        for i in range(g.shape[0]):
            xi_minus_X = (x[i] - X[X<=x[i]])/h
            g_num = np.sum(norm.pdf(xi_minus_X)*Y[X<=x[i]])
            g_den = np.sum(norm.pdf(xi_minus_X))

            g[i] = g_num/g_den

        return g

        # # Compute the difference between x and X for each point in x
        # xi_minus_X = (x[:,np.newaxis] - X)/h

        # # Compute the numerator and denominator of g(x)
        # g_num = np.sum(norm.pdf(xi_minus_X)*Y[np.newaxis,:],axis=1)
        # g_den = np.sum(norm.pdf(xi_minus_X),axis=1)

        # # Return g(x)
        # return g_num/g_den


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['py_np4vtt']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.5.1,<4.0.0',
 'numdifftools>=0.9.40,<0.10.0',
 'pandas>=1.3.1,<2.0.0',
 'scikit-learn>=1.0.2,<2.0.0',
 'scipy>=1.7.1,<2.0.0']

setup_kwargs = {
    'name': 'py-np4vtt',
    'version': '1.0.6',
    'description': 'Python library providing NonParametric models for Value of Travel Time analysis',
    'long_description': '# NP4VTT\n\nNP4VTT is a Python package that enables researchers to estimate and compare nonparametric models in a fast and convenient way. It comprises five nonparametric models for estimating the VTT distribution from data coming from two-attribute-two-alternative stated choice experiments:\n\n   * Local constant model  (Fosgerau, 2006, 2007)\n   * Local logit (Fosgerau, 2007)\n   * Rouwendal model (Rouwendal et al., 2010)\n   * Artificial Neural Network (ANN) based VTT model (van Cranenburgh & Kouwenhoven, 2021)\n   * Logistic Regression based VTT model (van Cranenburgh & Kouwenhoven, 2021)\n\nAdditionally, a Random Valuation model (Ojeda-Cabral, 2006) is included for benchmarking purposes\n\n## Installation steps\n\n* Use `pip` to install the `py-np4vtt` library normally:\n    - `python3 -m pip install py-np4vtt`\n\n\n## Examples\n\nWe provide Jupyter Notebooks that show how to configure and estimate each model included in NP4VTT:\n\n   * Local constant model: [link](https://gitlab.tudelft.nl/np4vtt/py-np4vtt/-/blob/master/examples/lconstant.ipynb)\n   * Local logit: [link](https://gitlab.tudelft.nl/np4vtt/py-np4vtt/-/blob/master/examples/loclogit.ipynb)\n   * Rouwendal model: [link](https://gitlab.tudelft.nl/np4vtt/py-np4vtt/-/blob/master/examples/rouwendal.ipynb)\n   * ANN-based VTT model: [link](https://gitlab.tudelft.nl/np4vtt/py-np4vtt/-/blob/master/examples/ann.ipynb)\n   * Logistic Regression-based VTT model: [link](https://gitlab.tudelft.nl/np4vtt/py-np4vtt/-/blob/master/examples/logistic.ipynb)\n\nThese examples guide the user through the process of loading a dataset, estimating a nonparametric model, and visualising the VTT distribution using scatter and histogram plots. We use the Norwegian 2009 VTT data to illustrate each example.\n\n**Take, for example, the VTT distribution from the Rouwendal model using NP4VTT:**\n\n![VTT distribution from the Rouwendal model using NP4VTT](examples/norway_data/outcomes/rouwendal.png)\n\n## References\n\n   * Fosgerau, M. (2006). Investigating the distribution of the value of travel time savings. Transportation Research Part B: Methodological, 40(8), 688–707. https://doi.org/10.1016/j.trb.2005.09.007\n   * Fosgerau, M. (2007). Using nonparametrics to specify a model to measure the value of travel time. Transportation Research Part A: Policy and Practice, 41(9), 842–856. https://doi.org/10.1016/j.tra.2006.10.004\n   * Rouwendal, J., de Blaeij, A., Rietveld, P., & Verhoef, E. (2010). The information content of a stated choice experiment: A new method and its application to the value of a statistical life. Transportation Research Part B: Methodological, 44(1), 136–151. https://doi.org/10.1016/j.trb.2009.04.006\n   * Ojeda-Cabral, M., Batley, R., & Hess, S. (2016). The value of travel time: Random utility versus random valuation. Transportmetrica A: Transport Science, 12(3), 230–248. https://doi.org/10.1080/23249935.2015.1125398\n   * van Cranenburgh, S., & Kouwenhoven, M. (2021). An artificial neural network based method to uncover the value-of-travel-time distribution. Transportation, 48(5), 2545–2583. https://doi.org/10.1007/s11116-020-10139-3',
    'author': 'José Ignacio Hernández',
    'author_email': 'J.I.Hernandez@tudelft.nl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.tudelft.nl/np4vtt/py-np4vtt',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<=3.10',
}


setup(**setup_kwargs)
